@@ -4,21 +4,51 @@
 //! STM32 microcontrollers. HALs for compatible chips can reexport this crate and implement its
 //! traits to easily expose a featureful CAN driver.
 //!
-//! Caveats:
-//! - Only RX FIFO 0 is supported, FIFO 1 will not be used.
+//! # Features
+//!
+//! - Supports both single- and dual-peripheral configurations (where one bxCAN instance manages the
+//!   filters of a secondary instance).
+//! - Handles standard and extended frames, and data and remote frames.
+//! - Support for interrupts emitted by the bxCAN peripheral.
+//! - Transmission respects CAN IDs and protects against priority inversion (a lower-priority frame
+//!   may be dequeued when enqueueing a higher-priority one).
+//! - Optionally implements the [`embedded-can`] traits for interoperability.
+//!
+//! # Limitations
+//!
+//! - Currently, only RX FIFO 0 is supported, and FIFO 1 will not be used.
+//! - Support for querying error states and handling error interrupts is incomplete.
+//!
+//! # Cargo Features
+//!
+//! | Feature | Description |
+//! |---------|-------------|
+//! | `unstable-defmt` | Implements [`defmt`]'s `Format` trait for the types in this crate.[^1] |
+//! | `embedded-can-03` | Implements the [`embedded-can`] 0.3 traits. |
+//!
+//! [^1]: The specific version of defmt is unspecified and may be updated in a patch release.
+//!
+//! [`embedded-can`]: https://docs.rs/embedded-can
+//! [`defmt`]: https://docs.rs/defmt
 
-#![doc(html_root_url = "https://docs.rs/bxcan/0.5.1")]
+#![doc(html_root_url = "https://docs.rs/bxcan/0.6.0")]
 // Deny a few warnings in doctests, since rustdoc `allow`s many warnings by default
 #![doc(test(attr(deny(unused_imports, unused_must_use))))]
 #![no_std]
+#![allow(clippy::unnecessary_operation)] // lint is bugged
 
+#[cfg(feature = "embedded-can-03")]
+mod embedded_can;
 pub mod filter;
 mod frame;
+mod id;
 mod interrupt;
-mod pac;
 mod readme;
 
-pub use embedded_can::{ExtendedId, Id, StandardId};
+#[allow(clippy::all)] // generated code
+mod pac;
+
+pub use id::{ExtendedId, Id, StandardId};
 
 pub use crate::frame::{Data, Frame, FramePriority};
 pub use crate::interrupt::{Interrupt, Interrupts};
@@ -28,8 +58,8 @@ use crate::filter::MasterFilters;
 use core::cmp::{Ord, Ordering};
 use core::convert::{Infallible, TryInto};
 use core::marker::PhantomData;
+use core::mem;
 use core::ptr::NonNull;
-use defmt::Format;
 
 use self::pac::generic::*; // To make the PAC extraction build
 
@@ -55,11 +85,11 @@ pub unsafe trait Instance {
 ///
 /// In master-slave-instance setups, only the master instance owns the filter banks, and needs to
 /// split some of them off for use by the slave instance. In that case, the master instance should
-/// implement `FilterOwner` and `MasterInstance`, while the slave instance should only implement
-/// `Instance`.
+/// implement [`FilterOwner`] and [`MasterInstance`], while the slave instance should only implement
+/// [`Instance`].
 ///
 /// In single-instance configurations, the instance owns all filter banks and they can not be split
-/// off. In that case, the instance should implement `Instance` and `FilterOwner`.
+/// off. In that case, the instance should implement [`Instance`] and [`FilterOwner`].
 ///
 /// # Safety
 ///
@@ -105,7 +135,8 @@ pub enum Error {
 /// Lower identifier values have a higher priority. Additionally standard frames
 /// have a higher priority than extended frames and data frames have a higher
 /// priority than remote frames.
-#[derive(Clone, Copy, Debug, PartialEq, Eq, Format)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
+#[cfg_attr(feature = "unstable-defmt", derive(defmt::Format))]
 struct IdReg(u32);
 
 impl IdReg {
@@ -210,16 +241,13 @@ impl PartialOrd for IdReg {
     }
 }
 
-/// Configuration proxy to be used with `Can::configure()`.
+/// Configuration proxy returned by [`Can::modify_config`].
+#[must_use = "`CanConfig` leaves the peripheral in uninitialized state, call `CanConfig::enable` or explicitly drop the value"]
 pub struct CanConfig<'a, I: Instance> {
-    _can: PhantomData<&'a mut I>,
+    can: &'a mut Can<I>,
 }
 
 impl<I: Instance> CanConfig<'_, I> {
-    fn registers(&self) -> &RegisterBlock {
-        unsafe { &*I::REGISTERS }
-    }
-
     /// Configures the bit timings.
     ///
     /// You can use <http://www.bittiming.can-wiki.info/> to calculate the `btr` parameter. Enter
@@ -232,8 +260,8 @@ impl<I: Instance> CanConfig<'_, I> {
     ///
     /// Then copy the `CAN_BUS_TIME` register value from the table and pass it as the `btr`
     /// parameter to this method.
-    pub fn set_bit_timing(&mut self, btr: u32) -> &mut Self {
-        let can = self.registers();
+    pub fn set_bit_timing(self, btr: u32) -> Self {
+        let can = self.can.registers();
         can.btr.modify(|r, w| unsafe {
             let mode_bits = r.bits() & 0xC000_0000;
             w.bits(mode_bits | btr)
@@ -243,25 +271,61 @@ impl<I: Instance> CanConfig<'_, I> {
 
     /// Enables or disables loopback mode: Internally connects the TX and RX
     /// signals together.
-    pub fn set_loopback(&mut self, enabled: bool) -> &mut Self {
-        let can = self.registers();
+    pub fn set_loopback(self, enabled: bool) -> Self {
+        let can = self.can.registers();
         can.btr.modify(|_, w| w.lbkm().bit(enabled));
         self
     }
 
     /// Enables or disables silent mode: Disconnects the TX signal from the pin.
-    pub fn set_silent(&mut self, enabled: bool) -> &mut Self {
-        let can = self.registers();
+    pub fn set_silent(self, enabled: bool) -> Self {
+        let can = self.can.registers();
         can.btr.modify(|_, w| w.silm().bit(enabled));
         self
     }
-}
 
-impl<I: Instance> Drop for CanConfig<'_, I> {
-    #[inline]
-    fn drop(&mut self) {
-        // Leave initialization mode
-        let can = self.registers();
+    /// Enables or disables automatic retransmission of messages.
+    ///
+    /// If this is enabled, the CAN peripheral will automatically try to retransmit each frame
+    /// until it can be sent. Otherwise, it will try only once to send each frame.
+    ///
+    /// Automatic retransmission is enabled by default.
+    pub fn set_automatic_retransmit(self, enabled: bool) -> Self {
+        let can = self.can.registers();
+        can.mcr.modify(|_, w| w.nart().bit(!enabled));
+        self
+    }
+
+    /// Leaves initialization mode and enables the peripheral.
+    ///
+    /// To sync with the CAN bus, this will block until 11 consecutive recessive bits are detected
+    /// on the bus.
+    ///
+    /// If you want to finish configuration without enabling the peripheral, you can call
+    /// [`CanConfig::leave_disabled`] or [`drop`] the [`CanConfig`] instead.
+    pub fn enable(mut self) {
+        self.leave_init_mode();
+
+        match nb::block!(self.can.enable_non_blocking()) {
+            Ok(()) => {}
+            Err(void) => match void {},
+        }
+
+        // Don't run the destructor.
+        mem::forget(self);
+    }
+
+    /// Leaves initialization mode, but keeps the peripheral in sleep mode.
+    ///
+    /// Before the [`Can`] instance can be used, you have to enable it by calling
+    /// [`Can::enable_non_blocking`].
+    pub fn leave_disabled(mut self) {
+        self.leave_init_mode();
+    }
+
+    /// Leaves initialization mode, enters sleep mode.
+    fn leave_init_mode(&mut self) {
+        let can = self.can.registers();
         can.mcr
             .modify(|_, w| w.sleep().set_bit().inrq().clear_bit());
         loop {
@@ -273,7 +337,111 @@ impl<I: Instance> Drop for CanConfig<'_, I> {
     }
 }
 
-/// Interface to a CAN peripheral.
+impl<I: Instance> Drop for CanConfig<'_, I> {
+    #[inline]
+    fn drop(&mut self) {
+        self.leave_init_mode();
+    }
+}
+
+/// Builder returned by [`Can::builder`].
+#[must_use = "`CanBuilder` leaves the peripheral in uninitialized state, call `CanBuilder::enable` or `CanBuilder::leave_disabled`"]
+pub struct CanBuilder<I: Instance> {
+    can: Can<I>,
+}
+
+impl<I: Instance> CanBuilder<I> {
+    /// Configures the bit timings.
+    ///
+    /// You can use <http://www.bittiming.can-wiki.info/> to calculate the `btr` parameter. Enter
+    /// parameters as follows:
+    ///
+    /// - *Clock Rate*: The input clock speed to the CAN peripheral (*not* the CPU clock speed).
+    ///   This is the clock rate of the peripheral bus the CAN peripheral is attached to (eg. APB1).
+    /// - *Sample Point*: Should normally be left at the default value of 87.5%.
+    /// - *SJW*: Should normally be left at the default value of 1.
+    ///
+    /// Then copy the `CAN_BUS_TIME` register value from the table and pass it as the `btr`
+    /// parameter to this method.
+    pub fn set_bit_timing(self, btr: u32) -> Self {
+        let can = self.can.registers();
+        can.btr.modify(|r, w| unsafe {
+            let mode_bits = r.bits() & 0xC000_0000;
+            w.bits(mode_bits | btr)
+        });
+        self
+    }
+
+    /// Enables or disables loopback mode: Internally connects the TX and RX
+    /// signals together.
+    pub fn set_loopback(self, enabled: bool) -> Self {
+        let can = self.can.registers();
+        can.btr.modify(|_, w| w.lbkm().bit(enabled));
+        self
+    }
+
+    /// Enables or disables silent mode: Disconnects the TX signal from the pin.
+    pub fn set_silent(self, enabled: bool) -> Self {
+        let can = self.can.registers();
+        can.btr.modify(|_, w| w.silm().bit(enabled));
+        self
+    }
+
+    /// Enables or disables automatic retransmission of messages.
+    ///
+    /// If this is enabled, the CAN peripheral will automatically try to retransmit each frame
+    /// until it can be sent. Otherwise, it will try only once to send each frame.
+    ///
+    /// Automatic retransmission is enabled by default.
+    pub fn set_automatic_retransmit(self, enabled: bool) -> Self {
+        let can = self.can.registers();
+        can.mcr.modify(|_, w| w.nart().bit(!enabled));
+        self
+    }
+
+    /// Leaves initialization mode and enables the peripheral.
+    ///
+    /// To sync with the CAN bus, this will block until 11 consecutive recessive bits are detected
+    /// on the bus.
+    ///
+    /// If you want to finish configuration without enabling the peripheral, you can call
+    /// [`CanBuilder::leave_disabled`] instead.
+    pub fn enable(mut self) -> Can<I> {
+        self.leave_init_mode();
+
+        match nb::block!(self.can.enable_non_blocking()) {
+            Ok(()) => self.can,
+            Err(void) => match void {},
+        }
+    }
+
+    /// Returns the [`Can`] interface without enabling it.
+    ///
+    /// This leaves initialization mode, but keeps the peripheral in sleep mode instead of enabling
+    /// it.
+    ///
+    /// Before the [`Can`] instance can be used, you have to enable it by calling
+    /// [`Can::enable_non_blocking`].
+    pub fn leave_disabled(mut self) -> Can<I> {
+        self.leave_init_mode();
+        self.can
+    }
+
+    /// Leaves initialization mode, enters sleep mode.
+    fn leave_init_mode(&mut self) {
+        let can = self.can.registers();
+        can.mcr
+            .modify(|_, w| w.sleep().set_bit().inrq().clear_bit());
+        loop {
+            let msr = can.msr.read();
+            if msr.slak().bit_is_set() && msr.inak().bit_is_clear() {
+                break;
+            }
+        }
+    }
+}
+
+/// Interface to a bxCAN peripheral.
 pub struct Can<I: Instance> {
     instance: I,
 }
@@ -282,9 +450,11 @@ impl<I> Can<I>
 where
     I: Instance,
 {
-    /// Creates a CAN interface, taking ownership of the raw peripheral.
-    pub fn new(instance: I) -> Self {
-        Can { instance }
+    /// Creates a [`CanBuilder`] for constructing a CAN interface.
+    pub fn builder(instance: I) -> CanBuilder<I> {
+        CanBuilder {
+            can: Can { instance },
+        }
     }
 
     fn registers(&self) -> &RegisterBlock {
@@ -309,32 +479,7 @@ where
 
     /// Configure bit timings and silent/loop-back mode.
     ///
-    /// Acutal configuration happens on the `CanConfig` that is passed to the
-    /// closure. It must be done this way because those configuration bits can
-    /// only be set if the CAN controller is in a special init mode.
-    /// Puts the peripheral in sleep mode afterwards. `Can::enable()` must be
-    /// called to exit sleep mode and start reception and transmission.
-    pub fn configure<F>(&mut self, f: F)
-    where
-        F: FnOnce(&mut CanConfig<I>),
-    {
-        let can = self.registers();
-
-        // Enter init mode.
-        can.mcr
-            .modify(|_, w| w.sleep().clear_bit().inrq().set_bit());
-        loop {
-            let msr = can.msr.read();
-            if msr.slak().bit_is_clear() && msr.inak().bit_is_set() {
-                break;
-            }
-        }
-
-        let mut config = CanConfig { _can: PhantomData };
-        f(&mut config);
-    }
-
-    /// Configure bit timings and silent/loop-back mode.
+    /// Calling this method will enter initialization mode.
     pub fn modify_config(&mut self) -> CanConfig<'_, I> {
         let can = self.registers();
 
@@ -348,19 +493,30 @@ where
             }
         }
 
-        CanConfig { _can: PhantomData }
+        CanConfig { can: self }
     }
 
     /// Configures the automatic wake-up feature.
+    ///
+    /// This is turned off by default.
+    ///
+    /// When turned on, an incoming frame will cause the peripheral to wake up from sleep and
+    /// receive the frame. If enabled, [`Interrupt::Wakeup`] will also be triggered by the incoming
+    /// frame.
     pub fn set_automatic_wakeup(&mut self, enabled: bool) {
         let can = self.registers();
         can.mcr.modify(|_, w| w.awum().bit(enabled));
     }
 
-    /// Start reception and transmission.
+    /// Leaves initialization mode and enables the peripheral (non-blocking version).
     ///
-    /// Waits for 11 consecutive recessive bits to sync to the CAN bus.
-    pub fn enable(&mut self) -> nb::Result<(), Infallible> {
+    /// Usually, it is recommended to call [`CanConfig::enable`] instead. This method is only needed
+    /// if you want non-blocking initialization.
+    ///
+    /// If this returns [`WouldBlock`][nb::Error::WouldBlock], the peripheral will enable itself
+    /// in the background. The peripheral is enabled and ready to use when this method returns
+    /// successfully.
+    pub fn enable_non_blocking(&mut self) -> nb::Result<(), Infallible> {
         let can = self.registers();
         let msr = can.msr.read();
         if msr.slak().bit_is_set() {
@@ -428,14 +584,16 @@ where
     }
 
     /// Clears the pending flag of [`Interrupt::Sleep`].
-    pub fn clear_sleep_interrupt(&mut self) {
+    pub fn clear_sleep_interrupt(&self) {
         let can = self.registers();
+        // Read-only register with write-1-to-clear, so `&self` is sufficient.
         can.msr.write(|w| w.slaki().set_bit());
     }
 
     /// Clears the pending flag of [`Interrupt::Wakeup`].
-    pub fn clear_wakeup_interrupt(&mut self) {
+    pub fn clear_wakeup_interrupt(&self) {
         let can = self.registers();
+        // Read-only register with write-1-to-clear, so `&self` is sufficient.
         can.msr.write(|w| w.wkui().set_bit());
     }
 
@@ -477,37 +635,24 @@ where
     /// Transmit order is preserved for frames with identical identifiers.
     /// If all transmit mailboxes are full, a higher priority frame replaces the
     /// lowest priority frame, which is returned as `Ok(Some(frame))`.
-    pub fn transmit(&mut self, frame: &Frame) -> nb::Result<Option<Frame>, Infallible> {
+    pub fn transmit(&mut self, frame: &Frame) -> nb::Result<TransmitStatus, Infallible> {
         // Safety: We have a `&mut self` and have unique access to the peripheral.
         unsafe { Tx::<I>::conjure().transmit(frame) }
     }
 
-    /// Puts a CAN frame in a free transmit mailbox for transmission on the bus.
-    ///
-    /// This function is equivalent to [`transmit`](#method.transmit) except that it returns the
-    /// mailbox that was accessed. This can be used to keep track of additional information
-    /// about each frame, even when frames are placed into transmit mailboxes and later
-    /// removed before being transmitted.
-    pub fn transmit_and_get_mailbox(
-        &mut self,
-        frame: &Frame,
-    ) -> nb::Result<(Option<Frame>, Mailbox), Infallible> {
-        // Safety: We have a `&mut self` and have unique access to the peripheral.
-        unsafe { Tx::<I>::conjure().transmit_and_get_mailbox(frame) }
-    }
-
     /// Returns `true` if no frame is pending for transmission.
     pub fn is_transmitter_idle(&self) -> bool {
-        // Safety: We have a `&mut self` and have unique access to the peripheral.
+        // Safety: Read-only operation.
         unsafe { Tx::<I>::conjure().is_idle() }
     }
 
     /// Attempts to abort the sending of a frame that is pending in a mailbox.
     ///
-    /// If there is no frame in the provided mailbox, this function has no effect and returns false.
+    /// If there is no frame in the provided mailbox, or its transmission succeeds before it can be
+    /// aborted, this function has no effect and returns `false`.
     ///
-    /// If there is a frame in the provided mailbox, this function returns true if the transmission
-    /// was aborted. This function returns false if the frame was successfully transmitted.
+    /// If there is a frame in the provided mailbox, and it is canceled successfully, this function
+    /// returns `true`.
     pub fn abort(&mut self, mailbox: Mailbox) -> bool {
         // Safety: We have a `&mut self` and have unique access to the peripheral.
         unsafe { Tx::<I>::conjure().abort(mailbox) }
@@ -545,26 +690,6 @@ impl<I: FilterOwner> Can<I> {
     }
 }
 
-impl<I> embedded_can::Can for Can<I>
-where
-    I: Instance,
-{
-    type Frame = Frame;
-
-    type Error = ();
-
-    fn try_transmit(
-        &mut self,
-        frame: &Self::Frame,
-    ) -> nb::Result<Option<Self::Frame>, Self::Error> {
-        self.transmit(frame).map_err(|e| e.map(|_| ()))
-    }
-
-    fn try_receive(&mut self) -> nb::Result<Self::Frame, Self::Error> {
-        self.receive()
-    }
-}
-
 /// Interface to the CAN transmitter part.
 pub struct Tx<I> {
     _can: PhantomData<I>,
@@ -590,7 +715,7 @@ where
 
     /// Creates a `&mut Self` out of thin air.
     ///
-    /// This is only safe if it is the only way to access an `Rx<I>`.
+    /// This is only safe if it is the only way to access a `Tx<I>`.
     unsafe fn conjure_by_ref<'a>() -> &'a mut Self {
         // Cause out of bounds access when `Self` is not zero-sized.
         [()][core::mem::size_of::<Self>()];
@@ -603,27 +728,14 @@ where
         unsafe { &*I::REGISTERS }
     }
 
-    /// Puts a CAN frame in a free transmit mailbox for transmission on the bus.
+    /// Puts a CAN frame in a transmit mailbox for transmission on the bus.
     ///
-    /// Frames are transmitted to the bus based on their priority (identifier).
-    /// Transmit order is preserved for frames with identical identifiers.
-    /// If all transmit mailboxes are full, a higher priority frame replaces the
-    /// lowest priority frame, which is returned as `Ok(Some(frame))`.
-    pub fn transmit(&mut self, frame: &Frame) -> nb::Result<Option<Frame>, Infallible> {
-        self.transmit_and_get_mailbox(frame)
-            .map(|(frame, _mailbox)| frame)
-    }
-
-    /// Puts a CAN frame in a free transmit mailbox for transmission on the bus.
+    /// Frames are transmitted to the bus based on their priority (identifier). Transmit order is
+    /// preserved for frames with identical identifiers.
     ///
-    /// This function is equivalent to [`transmit`](#method.transmit) except that it returns the
-    /// mailbox that was accessed. This can be used to keep track of additional information
-    /// about each frame, even when frames are placed into transmit mailboxes and later
-    /// removed before being transmitted.
-    pub fn transmit_and_get_mailbox(
-        &mut self,
-        frame: &Frame,
-    ) -> nb::Result<(Option<Frame>, Mailbox), Infallible> {
+    /// If all transmit mailboxes are full, a higher priority frame can replace a lower-priority
+    /// frame, which is returned in the [`TransmitStatus`].
+    pub fn transmit(&mut self, frame: &Frame) -> nb::Result<TransmitStatus, Infallible> {
         let can = self.registers();
 
         // Get the index of the next free mailbox or the one with the lowest priority.
@@ -667,7 +779,10 @@ where
             2 => Mailbox::Mailbox2,
             _ => unreachable!(),
         };
-        Ok((pending_frame, mailbox))
+        Ok(TransmitStatus {
+            dequeued_frame: pending_frame,
+            mailbox,
+        })
     }
 
     /// Returns `Ok` when the mailbox is free or if it contains pending frame with a
@@ -732,7 +847,7 @@ where
             // Abort request failed because the frame was already sent (or being sent) on
             // the bus. All mailboxes are now free. This can happen for small prescaler
             // values (e.g. 1MBit/s bit timing with a source clock of 8MHz) or when an ISR
-            // has preemted the execution.
+            // has preempted the execution.
             None
         }
     }
@@ -754,10 +869,11 @@ where
 
     /// Attempts to abort the sending of a frame that is pending in a mailbox.
     ///
-    /// If there is no frame in the provided mailbox, this function has no effect and returns false.
+    /// If there is no frame in the provided mailbox, or its transmission succeeds before it can be
+    /// aborted, this function has no effect and returns `false`.
     ///
-    /// If there is a frame in the provided mailbox, this function returns true if the transmission
-    /// was aborted. This function returns false if the frame was successfully transmitted.
+    /// If there is a frame in the provided mailbox, and it is canceled successfully, this function
+    /// returns `true`.
     pub fn abort(&mut self, mailbox: Mailbox) -> bool {
         // If the mailbox is empty, the value of TXOKx depends on what happened with the previous
         // frame in that mailbox. Only call abort_by_index() if the mailbox is not empty.
@@ -862,8 +978,9 @@ where
     }
 }
 
-/// The three transmit mailboxes
-#[derive(Debug, Clone, Ord, PartialOrd, Eq, PartialEq, Format)]
+/// The three transmit mailboxes.
+#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq)]
+#[cfg_attr(feature = "unstable-defmt", derive(defmt::Format))]
 pub enum Mailbox {
     /// Transmit mailbox 0
     Mailbox0 = 0,
@@ -872,3 +989,24 @@ pub enum Mailbox {
     /// Transmit mailbox 2
     Mailbox2 = 2,
 }
+
+/// Contains information about a frame enqueued for transmission via [`Can::transmit`] or
+/// [`Tx::transmit`].
+pub struct TransmitStatus {
+    dequeued_frame: Option<Frame>,
+    mailbox: Mailbox,
+}
+
+impl TransmitStatus {
+    /// Returns the lower-priority frame that was dequeued to make space for the new frame.
+    #[inline]
+    pub fn dequeued_frame(&self) -> Option<&Frame> {
+        self.dequeued_frame.as_ref()
+    }
+
+    /// Returns the [`Mailbox`] the frame was enqueued in.
+    #[inline]
+    pub fn mailbox(&self) -> Mailbox {
+        self.mailbox
+    }
+}
